#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .msk_cluster import MskCluster
    from ecs_composex.common.settings import ComposeXSettings
    from troposphere import Template

from compose_x_common.compose_x_common import keyisset
from ecs_composex.common import get_nested_property
from ecs_composex.common.logging import LOG
from ecs_composex.common.troposphere_tools import add_resource
from ecs_composex.kinesis_firehose.kinesis_firehose_stack import DeliveryStream
from ecs_composex.s3.s3_bucket import Bucket
from troposphere import Ref
from troposphere.logs import LogGroup
from troposphere.msk import BrokerLogs, CloudWatchLogs, LoggingInfo


def add_log_group(cluster: MskCluster, template: Template) -> None:
    """
    Creates a new Log Group for the cluster and configures the property for it.
    """
    resource_props_to_update: list = get_nested_property(
        cluster.cfn_resource, "LoggingInfo.BrokerLogs.CloudWatchLogs"
    )
    for _resource_prop_to_update in resource_props_to_update:
        param, prop, value = _resource_prop_to_update
        if isinstance(value, CloudWatchLogs):
            LOG.warning(
                f"{cluster.module.res_key}.{cluster.name} - "
                "Properties for LoggingInfo.BrokerLogs.CloudWatchLogs already set. Skipping"
            )
            return
        log_group = LogGroup(f"{cluster.logical_name}BrokersLogGroup")
        add_resource(template, log_group)
        if value and value is not None:
            setattr(param, prop, CloudWatchLogs(Enabled=True, LogGroup=Ref(log_group)))
        else:
            if not hasattr(cluster.cfn_resource, "LoggingInfo"):
                setattr(
                    cluster.cfn_resource,
                    "LoggingInfo",
                    LoggingInfo(
                        BrokerLogs=BrokerLogs(
                            CloudWatchLogs=CloudWatchLogs(
                                Enabled=True, LogGroup=Ref(log_group)
                            )
                        )
                    ),
                )
            else:
                if not hasattr(
                    cluster.cfn_resource.LoggingInfo.BrokerLogs, "CloudWatchLogs"
                ):
                    setattr(
                        cluster.cfn_resource.LoggingInfo,
                        "CloudWatchLogs",
                        CloudWatchLogs(Enabled=True, LogGroup=Ref(log_group)),
                    )


def handle_logging(
    cluster: MskCluster, cluster_template: Template, settings: ComposeXSettings
) -> None:
    if cluster.parameters and keyisset("CreateLogGroup", cluster.parameters):
        add_log_group(cluster, cluster_template)



#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>


from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ecs_composex.common.settings import ComposeXSettings
    from .msk_cluster import MskCluster

from compose_x_common.compose_x_common import keyisset, set_else_none
from ecs_composex.common.logging import LOG

from .msk_cluster_ecs_iam import handle_iam_kafka_resources_access


def handle_kafka_iam_permissions(resource: MskCluster, settings: ComposeXSettings):
    for target in resource.families_targets:
        if target[0] and (not target[0].stack or not target[0].template):
            continue
        if not target[3]:
            LOG.warning(
                f"{resource.module.res_key}.{resource.name} - Access not defined for {target[0].name}"
            )
            continue
        kafka_access = set_else_none("KafkaAccess", target[-1])
        if keyisset("Iam", kafka_access):
            handle_iam_kafka_resources_access(
                resource, settings, target[0], kafka_access["Iam"]
            )



#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>

"""
MSK Configuration
"""

from __future__ import annotations

from typing import Any, Optional

from troposphere.msk import Configuration


class MskConfiguration(Configuration):
    default_configurations: dict = {
        "auto.create.topics.enable": False,
        "default.replication.factor": 3,
        "min.insync.replicas": 2,
        "num.io.threads": 8,
        "num.network.threads": 5,
        "num.partitions": 1,
        "num.replica.fetchers": 2,
        "replica.lag.time.max.ms": 30000,
        "socket.receive.buffer.bytes": 102400,
        "socket.request.max.bytes": 104857600,
        "socket.send.buffer.bytes": 102400,
        "unclean.leader.election.enable": True,
        "zookeeper.session.timeout.ms": 18000,
    }

    def __init__(self, title: Optional[str], **kwargs: Any):
        self.configuration_dict: dict = {}
        if "ServerProperties" not in kwargs:
            kwargs.update(
                {
                    "ServerProperties": self.config_dict_to_string(
                        self.default_configurations
                    )
                }
            )
        super().__init__(title, **kwargs)

    @staticmethod
    def config_dict_to_string(config: dict) -> str:
        return "\n".join(["%s=%s" % (key, value) for key, value in config.items()])



# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2022 John Mille <john@compose-x.io>

"""
Configure MSK Cluster autoscaling
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from troposphere import Template
    from .msk_cluster import MskCluster

from compose_x_common.compose_x_common import set_else_none
from ecs_composex.common.cfn_conditions import define_stack_name
from ecs_composex.common.troposphere_tools import add_resource
from troposphere import (
    AWS_ACCOUNT_ID,
    AWS_PARTITION,
    AWS_REGION,
    AWS_URL_SUFFIX,
    Ref,
    Sub,
)
from troposphere.applicationautoscaling import (
    PredefinedMetricSpecification,
    ScalableTarget,
    ScalableTargetAction,
    ScalingPolicy,
    TargetTrackingScalingPolicyConfiguration,
)


def set_scalable_target(
    cluster: MskCluster, storage_scaling: dict
) -> Union[None, ScalableTarget]:
    return ScalableTarget(
        f"{cluster.logical_name}ScalableTarget",
        DependsOn=[cluster.cfn_resource.title],
        ServiceNamespace="kafka",
        MaxCapacity=storage_scaling["MaxInGB"],
        MinCapacity=1,
        ResourceId=Ref(cluster.cfn_resource),
        RoleARN=Sub(
            f"arn:${{{AWS_PARTITION}}}:iam::${{{AWS_ACCOUNT_ID}}}:role/"
            f"aws-service-role/kafka.application-autoscaling.${{{AWS_URL_SUFFIX}}}/"
            "AWSServiceRoleForApplicationAutoScaling_KafkaCluster"
        ),
        ScalableDimension="kafka:broker-storage:VolumeSize",
    )


def set_scaling_policy(
    cluster: MskCluster,
    storage_scaling: dict,
    scaling_target: ScalableTarget,
) -> ScalingPolicy:
    return ScalingPolicy(
        f"{cluster.logical_name}StorageScalingPolicy",
        DependsOn=[scaling_target.title],
        PolicyName=Sub(
            "MSK Storage Scaling - ${StackName}", StackName=define_stack_name()
        ),
        PolicyType="TargetTrackingScaling",
        ScalingTargetId=Ref(scaling_target),
        TargetTrackingScalingPolicyConfiguration=TargetTrackingScalingPolicyConfiguration(
            DisableScaleIn=True,
            TargetValue=float(storage_scaling["Target"]),
            PredefinedMetricSpecification=PredefinedMetricSpecification(
                PredefinedMetricType="KafkaBrokerStorageUtilization"
            ),
        ),
    )


def add_storage_scaling(cluster: MskCluster, template: Template) -> None:
    """
    Adds storage autoscaling to the MSK Cluster. Application Autoscaling will add capacity
    as the data on the cluster grows. However, it won't scale in (remove capacity).
    This feature is mostly aimed at ensuring there is always storage available on the Kafka cluster
    """

    if not cluster.cfn_resource or not cluster.parameters:
        return
    storage_scaling = set_else_none("StorageScaling", cluster.parameters)
    if not storage_scaling:
        return
    scaling_target = set_scalable_target(cluster, storage_scaling)
    scaling_policy = set_scaling_policy(cluster, storage_scaling, scaling_target)
    add_resource(template, scaling_target)
    add_resource(template, scaling_policy)



#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ecs_composex.mods_manager import XResourceModule
    from ecs_composex.common.settings import ComposeXSettings
    from ecs_composex.mods_manager import ModManager

from compose_x_common.aws.msk import MSK_CLUSTER_ARN_RE
from compose_x_common.compose_x_common import keyisset
from ecs_composex.common.logging import LOG
from ecs_composex.compose.x_resources.network_x_resources import NetworkXResource
from ecs_composex.ingress_settings import lookup_security_group
from ecs_composex.resource_settings import link_resource_to_services
from ecs_composex.vpc.vpc_params import STORAGE_SUBNETS
from troposphere import GetAtt, Ref, Select, Split
from troposphere.msk import Cluster as CfnMskCluster

from ecs_composex_msk_cluster.msk_cluster_params import (
    CONTROL_CLOUD_ATTR_MAPPING,
    MSK_CLUSTER_ARN,
    MSK_CLUSTER_CLIENTS_SHARED_SG,
    MSK_CLUSTER_SG_PARAM,
)

from .msk_cluster_ecs import handle_kafka_iam_permissions


class MskCluster(NetworkXResource):
    """
    Class to manage MSK Cluster resource
    """

    def __init__(
        self,
        name: str,
        definition: dict,
        module: XResourceModule,
        settings: ComposeXSettings,
    ):
        super().__init__(name, definition, module, settings)

        self.clients_security_group = None
        self.security_group_param = MSK_CLUSTER_SG_PARAM
        self.clients_security_group_param = MSK_CLUSTER_CLIENTS_SHARED_SG
        self.subnets_override = STORAGE_SUBNETS
        self.post_processing_properties: list = [
            "EncryptionInfo.EncryptionAtRest.DataVolumeKMSKeyId",
            "BrokerNodeGroupInfo.ClientSubnets",
        ]
        self.ref_parameter = MSK_CLUSTER_ARN
        self.cluster_arn_parameter = MSK_CLUSTER_ARN
        self.cloud_control_attributes_mapping = CONTROL_CLOUD_ATTR_MAPPING

    @property
    def cluster_uuid(self):
        if self.cfn_resource:
            return Select(1, Split(":cluster/", Ref(self.cfn_resource)))
        else:
            return Select(
                1,
                Split(
                    ":cluster/", self.attributes_outputs[MSK_CLUSTER_ARN]["ImportValue"]
                ),
            )

    def init_outputs(self):
        self.output_properties: dict = {
            MSK_CLUSTER_ARN: (self.logical_name, self.cfn_resource, Ref, None),
            MSK_CLUSTER_SG_PARAM: (
                f"{self.logical_name}{self.security_group_param.title}",
                self.security_group,
                GetAtt,
                self.security_group_param.return_value,
            ),
            MSK_CLUSTER_CLIENTS_SHARED_SG: (
                f"{self.logical_name}{self.clients_security_group_param.title}",
                self.clients_security_group,
                GetAtt,
                self.clients_security_group_param.return_value,
            ),
        }

    def to_ecs(self, settings, modules: ModManager, root_stack=None) -> None:
        """
        Maps a database service to ECS services
        """
        for target in self.families_targets:
            if target[0].service_compute.launch_type != "EXTERNAL":
                if self.cfn_resource:
                    client_sg_id = self.add_attribute_to_another_stack(
                        target[0].stack, self.clients_security_group_param, settings
                    )
                    target[0].service_networking.extra_security_groups.append(
                        client_sg_id["ImportParameter"]
                    )
                    target[0].ecs_service.ecs_service.NetworkConfiguration = target[
                        0
                    ].service_networking.ecs_network_config
                elif (
                    self.lookup_properties
                    and MSK_CLUSTER_CLIENTS_SHARED_SG in self.lookup_properties
                    and MSK_CLUSTER_CLIENTS_SHARED_SG in self.attributes_outputs
                ):
                    LOG.info(
                        f"{self.module.res_key}.{self.name} - "
                        "Associating found ClientsSecurityGroup from Lookup."
                    )
                    target[0].service_networking.extra_security_groups.append(
                        self.attributes_outputs[MSK_CLUSTER_CLIENTS_SHARED_SG][
                            "ImportValue"
                        ]
                    )
                    target[0].ecs_service.ecs_service.NetworkConfiguration = target[
                        0
                    ].service_networking.ecs_network_config
            else:
                LOG.warning(
                    f"{self.stack.title} - {target[0].name} - "
                    "When using EXTERNAL Launch Type, networking settings cannot be set."
                )
        if self.cluster_arn_parameter:
            link_resource_to_services(
                settings,
                self,
                arn_parameter=self.cluster_arn_parameter,
                access_subkeys=["MSKCluster"],
            )
            handle_kafka_iam_permissions(self, settings)


def describe_kafka_cluster(
    cluster: MskCluster, account_id: str, resource_id: str
) -> dict:
    client = cluster.lookup_session.client("kafka")
    cluster_r = client.describe_cluster_v2(ClusterArn=resource_id)
    return cluster_r["ClusterInfo"]


def define_msk_clusters_mappings(module: XResourceModule, settings: ComposeXSettings):
    for cluster in module.lookup_resources:
        cluster.init_outputs()
        cluster.lookup_resource(
            MSK_CLUSTER_ARN_RE,
            describe_kafka_cluster,
            CfnMskCluster.resource_type,
            "kafka:cluster",
            "Cluster",
            use_arn_for_id=True,
        )
        if keyisset("ClientsSecurityGroup", cluster.lookup):
            client_sg = lookup_security_group(
                settings, cluster.lookup["ClientsSecurityGroup"]
            )
            cluster.lookup_properties.update({MSK_CLUSTER_CLIENTS_SHARED_SG: client_sg})
            cluster.generate_cfn_mappings_from_lookup_properties()
            cluster.generate_outputs()
        else:
            LOG.warning(
                f"{module.res_key}.{cluster.name} - "
                "ClientsSecurityGroup not set. Inbound to MSK might not work."
            )

        settings.mappings[module.mapping_key].update(
            {cluster.logical_name: cluster.mappings}
        )



#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from troposphere import Template

from troposphere import Condition, Equals, If, Not, Or, Ref

from .msk_cluster_params import (
    MSK_CLUSTER_ADDRESSING_TYPE,
    MSK_CLUSTER_USE_SASL_IAM,
    MSK_CLUSTER_USE_SASL_SCRAM,
    MSK_CLUSTER_USE_TLS_AUTH,
)

CLUSTER_PUBLICLY_ADDRESSED_CON_T = "MskClusterPubliclyAddressed"
CLUSTER_PUBLICLY_ADDRESSED_CON = Equals(Ref(MSK_CLUSTER_ADDRESSING_TYPE), "PUBLIC")

CLUSTER_PRIVATE_ONLY_CON_T = "MskClusterPrivateOnly"
CLUSTER_PRIVATE_ONLY_CON = Not(Condition(CLUSTER_PUBLICLY_ADDRESSED_CON_T))

CLIENTS_USE_SASL_IAM_CON_T = "UseSaslIamCondition"
CLIENTS_USE_SASL_IAM_CON = Equals(Ref(MSK_CLUSTER_USE_SASL_IAM), "True")

CLIENTS_USE_SASL_SCRAM_CON_T = "UseSaslIamCondition"
CLIENTS_USE_SASL_SCRAM_CON = Equals(Ref(MSK_CLUSTER_USE_SASL_SCRAM), "True")

CLIENTS_USE_TLS_AUTH_CON_T = "UseTlsAuthCondition"
CLIENTS_USE_TLS_AUTH_CON = Equals(Ref(MSK_CLUSTER_USE_TLS_AUTH), "True")



#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>

from __future__ import annotations

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from .msk_cluster import MskCluster

from ecs_composex.common import get_nested_property
from troposphere import AccountId, FindInMap, GetAtt, If, NoValue, Ref

from .msk_cluster_conditions import (
    CLIENTS_USE_SASL_IAM_CON_T,
    CLIENTS_USE_SASL_SCRAM_CON_T,
    CLIENTS_USE_TLS_AUTH_CON_T,
    CLUSTER_PUBLICLY_ADDRESSED_CON_T,
)
from .msk_cluster_params import (
    MSK_CLUSTER_ADDRESSING_TYPE,
    MSK_CLUSTER_USE_SASL_IAM,
    MSK_CLUSTER_USE_SASL_SCRAM,
    MSK_CLUSTER_USE_TLS_AUTH,
    MSK_PORTS_MAPPING,
)


def str_to_bool(input_var: Union[str, bool]) -> bool:
    if isinstance(input_var, bool):
        return input_var
    elif isinstance(input_var, str):
        if input_var.lower() in ["true", "yes"]:
            return True
        return False
    return False


def allow_clients_sg_ingress(cluster) -> list[If]:
    rules: list = []

    for port_type, port in MSK_PORTS_MAPPING["Zookeeper"].items():
        rules.append(
            {
                "Description": f"Zookeeper {port_type} MSK",
                "FromPort": port,
                "ToPort": port,
                "SourceSecurityGroupId": GetAtt(
                    cluster.clients_security_group, "GroupId"
                ),
                "SourceSecurityGroupOwnerId": AccountId,
                "IpProtocol": 6,
            }
        )
    rules.append(
        If(
            CLIENTS_USE_SASL_IAM_CON_T,
            If(
                CLUSTER_PUBLICLY_ADDRESSED_CON_T,
                [
                    {
                        "Description": f"SASL IAM Public Access",
                        "FromPort": FindInMap("MSKPorts", "Public", "Iam"),
                        "ToPort": FindInMap("MSKPorts", "Public", "Iam"),
                        "SourceSecurityGroupId": GetAtt(
                            cluster.clients_security_group, "GroupId"
                        ),
                        "SourceSecurityGroupOwnerId": AccountId,
                        "IpProtocol": 6,
                    },
                    {
                        "Description": f"SASL IAM Private Access",
                        "FromPort": FindInMap("MSKPorts", "Private", "Iam"),
                        "ToPort": FindInMap("MSKPorts", "Private", "Iam"),
                        "SourceSecurityGroupId": GetAtt(
                            cluster.clients_security_group, "GroupId"
                        ),
                        "SourceSecurityGroupOwnerId": AccountId,
                        "IpProtocol": 6,
                    },
                ],
                {
                    "Description": f"SASL IAM Private Access",
                    "FromPort": FindInMap("MSKPorts", "Private", "Iam"),
                    "ToPort": FindInMap("MSKPorts", "Private", "Iam"),
                    "SourceSecurityGroupId": GetAtt(
                        cluster.clients_security_group, "GroupId"
                    ),
                    "SourceSecurityGroupOwnerId": AccountId,
                    "IpProtocol": 6,
                },
            ),
            NoValue,
        )
    )
    rules.append(
        If(
            CLIENTS_USE_SASL_SCRAM_CON_T,
            If(
                CLUSTER_PUBLICLY_ADDRESSED_CON_T,
                [
                    {
                        "Description": f"SASL SCRAM Public Access",
                        "FromPort": FindInMap("MSKPorts", "Public", "Scram"),
                        "ToPort": FindInMap("MSKPorts", "Public", "Scram"),
                        "SourceSecurityGroupId": GetAtt(
                            cluster.clients_security_group, "GroupId"
                        ),
                        "SourceSecurityGroupOwnerId": AccountId,
                        "IpProtocol": 6,
                    },
                    {
                        "Description": f"SASL SCRAM Private Access",
                        "FromPort": FindInMap("MSKPorts", "Private", "Scram"),
                        "ToPort": FindInMap("MSKPorts", "Private", "Scram"),
                        "SourceSecurityGroupId": GetAtt(
                            cluster.clients_security_group, "GroupId"
                        ),
                        "SourceSecurityGroupOwnerId": AccountId,
                        "IpProtocol": 6,
                    },
                ],
                {
                    "Description": f"SASL SCRAM Private Access",
                    "FromPort": FindInMap("MSKPorts", "Private", "Scram"),
                    "ToPort": FindInMap("MSKPorts", "Private", "Scram"),
                    "SourceSecurityGroupId": GetAtt(
                        cluster.clients_security_group, "GroupId"
                    ),
                    "SourceSecurityGroupOwnerId": AccountId,
                    "IpProtocol": 6,
                },
            ),
            NoValue,
        )
    )
    rules.append(
        If(
            CLIENTS_USE_TLS_AUTH_CON_T,
            If(
                CLUSTER_PUBLICLY_ADDRESSED_CON_T,
                [
                    {
                        "Description": f"TLS Authentication Public Access",
                        "FromPort": FindInMap("MSKPorts", "Public", "Tls"),
                        "ToPort": FindInMap("MSKPorts", "Public", "Tls"),
                        "SourceSecurityGroupId": GetAtt(
                            cluster.clients_security_group, "GroupId"
                        ),
                        "SourceSecurityGroupOwnerId": AccountId,
                        "IpProtocol": 6,
                    },
                    {
                        "Description": f"TLS Authentication Private Access",
                        "FromPort": FindInMap("MSKPorts", "Private", "Tls"),
                        "ToPort": FindInMap("MSKPorts", "Private", "Tls"),
                        "SourceSecurityGroupId": GetAtt(
                            cluster.clients_security_group, "GroupId"
                        ),
                        "SourceSecurityGroupOwnerId": AccountId,
                        "IpProtocol": 6,
                    },
                ],
                {
                    "Description": f"TLS Authentication Private Access",
                    "FromPort": FindInMap("MSKPorts", "Private", "Tls"),
                    "ToPort": FindInMap("MSKPorts", "Private", "Tls"),
                    "SourceSecurityGroupId": GetAtt(
                        cluster.clients_security_group, "GroupId"
                    ),
                    "SourceSecurityGroupOwnerId": AccountId,
                    "IpProtocol": 6,
                },
            ),
            NoValue,
        )
    )
    return rules


def handle_msk_auth_parameters(cluster: MskCluster) -> None:
    properties: list = [
        (
            "ClientAuthentication.Sasl.Scram.Enabled",
            str_to_bool,
            MSK_CLUSTER_USE_SASL_SCRAM,
        ),
        (
            "ClientAuthentication.Sasl.Iam.Enabled",
            str_to_bool,
            MSK_CLUSTER_USE_SASL_IAM,
        ),
        (
            "ClientAuthentication.Tls.Enabled",
            str_to_bool,
            MSK_CLUSTER_USE_TLS_AUTH,
        ),
    ]
    for auth_property in properties:
        res_props_to_update = get_nested_property(
            cluster.cfn_resource,
            auth_property[0],
        )
        for _prop_to_update in res_props_to_update:
            resource, prop, value = _prop_to_update
            if prop and value:
                cluster.stack.Parameters.update(
                    {auth_property[2].title: str(auth_property[1](value))}
                )
                setattr(resource, prop, Ref(auth_property[2]))


def handle_msk_auth_settings(cluster: MskCluster):
    setattr(
        cluster.security_group,
        "SecurityGroupIngress",
        allow_clients_sg_ingress(cluster),
    )
    resource_props_to_update = get_nested_property(
        cluster.cfn_resource,
        "BrokerNodeGroupInfo.ConnectivityInfo.PublicAccess.Type",
    )
    for _to_update in resource_props_to_update:
        resource, prop, value = _to_update
        if value and value == "SERVICE_PROVIDED_EIPS":
            cluster.stack.Parameters.update(
                {MSK_CLUSTER_ADDRESSING_TYPE.title: "PUBLIC"}
            )
        handle_msk_auth_parameters(cluster)



#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from troposphere import Template
    from ecs_composex.mods_manager import XResourceModule
    from .msk_cluster import MskCluster

from compose_x_common.aws.msk import list_all_kafka_versions
from compose_x_common.compose_x_common import keyisset, set_else_none
from ecs_composex.common.cfn_conditions import define_stack_name
from ecs_composex.common.cfn_params import ROOT_STACK_NAME
from ecs_composex.common.logging import LOG
from ecs_composex.common.stacks import ComposeXStack
from ecs_composex.common.troposphere_tools import (
    Parameter,
    add_outputs,
    add_resource,
    build_template,
)
from ecs_composex.resources_import import import_record_properties
from ecs_composex.vpc.vpc_params import (
    APP_SUBNETS,
    PUBLIC_SUBNETS,
    STORAGE_SUBNETS,
    VPC_ID,
)
from troposphere import GetAtt, Output, Ref, Sub
from troposphere.ec2 import SecurityGroup
from troposphere.msk import Cluster as AwsMskCluster

from .msk_cluster_conditions import (
    CLIENTS_USE_SASL_IAM_CON,
    CLIENTS_USE_SASL_IAM_CON_T,
    CLIENTS_USE_SASL_SCRAM_CON,
    CLIENTS_USE_SASL_SCRAM_CON_T,
    CLIENTS_USE_TLS_AUTH_CON,
    CLIENTS_USE_TLS_AUTH_CON_T,
    CLUSTER_PRIVATE_ONLY_CON,
    CLUSTER_PRIVATE_ONLY_CON_T,
    CLUSTER_PUBLICLY_ADDRESSED_CON,
    CLUSTER_PUBLICLY_ADDRESSED_CON_T,
)
from .msk_cluster_params import (
    MSK_CLUSTER_ADDRESSING_TYPE,
    MSK_CLUSTER_INSTANCE_TYPES,
    MSK_CLUSTER_SG_PARAM,
    MSK_CLUSTER_USE_SASL_IAM,
    MSK_CLUSTER_USE_SASL_SCRAM,
    MSK_PORTS_MAPPING,
)
from .msk_sg_ingress import handle_msk_auth_settings
from .msk_storage_scaling import add_storage_scaling


class MskClusterStack(ComposeXStack):
    """
    Class to represent an MSK Cluster' Stack
    """

    def __init__(self, name, stack_template, cluster, top_stack, **kwargs):
        self.cluster = cluster
        super().__init__(name, stack_template, **kwargs)
        self.parent_stack = top_stack
        self.cluster.stack = self


def validate_cluster_version(cluster: MskCluster, input_version) -> Parameter:
    """
    Validates the kafka version
    """
    valid_versions: list[str] = [
        _version["Version"]
        for _version in list_all_kafka_versions(session=cluster.lookup_session)
        if _version["Status"] == "ACTIVE"
    ]
    if input_version in valid_versions:
        return Parameter(
            "KafkaVersion",
            Type="String",
            AllowedValues=valid_versions,
            Default=input_version,
        )
    else:
        raise ValueError(
            "Version {} is not valid. Active versions supported: {}".format(
                input_version, valid_versions
            )
        )


def set_msk_cluster_template(module: XResourceModule, cluster: MskCluster) -> Template:
    template = build_template(
        f"{module.res_key}.{cluster.name} Stack Template",
        [
            MSK_CLUSTER_USE_SASL_IAM,
            MSK_CLUSTER_USE_SASL_SCRAM,
            MSK_CLUSTER_ADDRESSING_TYPE,
            VPC_ID,
            STORAGE_SUBNETS,
            APP_SUBNETS,
            PUBLIC_SUBNETS,
            MSK_CLUSTER_INSTANCE_TYPES,
        ],
    )
    template.add_condition(
        CLUSTER_PUBLICLY_ADDRESSED_CON_T, CLUSTER_PUBLICLY_ADDRESSED_CON
    )
    template.add_condition(CLUSTER_PRIVATE_ONLY_CON_T, CLUSTER_PRIVATE_ONLY_CON)
    template.add_condition(CLIENTS_USE_SASL_SCRAM_CON_T, CLIENTS_USE_SASL_SCRAM_CON)
    template.add_condition(CLIENTS_USE_SASL_IAM_CON_T, CLIENTS_USE_SASL_IAM_CON)
    template.add_condition(CLIENTS_USE_TLS_AUTH_CON_T, CLIENTS_USE_TLS_AUTH_CON)
    template.add_mapping("MSKPorts", MSK_PORTS_MAPPING)
    return template


def set_instance_type(cluster: MskCluster, cluster_stack: ComposeXStack) -> None:
    """
    Checks the instance type value. Replaces it with parameter.
    Updates parameter value if valid, uses Default if not
    """
    instance_type = getattr(cluster.cfn_resource.BrokerNodeGroupInfo, "InstanceType")
    if instance_type not in MSK_CLUSTER_INSTANCE_TYPES.AllowedValues:
        LOG.error(
            "{}.{} - {} InstanceType is not valid. Using default {}".format(
                cluster.module.res_key,
                cluster.name,
                instance_type,
                MSK_CLUSTER_INSTANCE_TYPES.Default,
            )
        )
    else:
        cluster_stack.Parameters.update(
            {MSK_CLUSTER_INSTANCE_TYPES.title: instance_type}
        )
    setattr(
        cluster.cfn_resource.BrokerNodeGroupInfo,
        "InstanceType",
        Ref(MSK_CLUSTER_INSTANCE_TYPES),
    )


def build_msk_clusters(module: XResourceModule, msk_top_stack: ComposeXStack):
    """
    Creates a new MSK cluster from properties
    Each MSK cluster gets its own template & own stack to allow for better re-usability.
    """
    for cluster in module.new_resources:
        cluster_template = set_msk_cluster_template(module, cluster)
        if cluster.parameters and keyisset(
            "CreateKafkaConfiguration", cluster.parameters
        ):
            LOG.warning(
                f"{cluster.module.res_key}.{cluster.name} - "
                "At the moment, the AWS::MSK::Configuration does not return the Revision. "
                "Therefore cannot be used here yet. "
                "See https://github.com/aws-cloudformation/cloudformation-coverage-roadmap/issues/1138"
            )
        cluster.clients_security_group = add_resource(
            cluster_template,
            SecurityGroup(
                f"{cluster.logical_name}ClientsSecurityGroup",
                GroupName=Sub(
                    f"${{StackName}}-{cluster.logical_name}--clients",
                    StackName=define_stack_name(cluster_template),
                ),
                GroupDescription=Sub(
                    f"${{StackName}} {cluster.name} Clients SG",
                    StackName=define_stack_name(),
                ),
                VpcId=Ref(VPC_ID),
            ),
        )
        cluster.security_group = add_resource(
            cluster_template,
            SecurityGroup(
                f"{cluster.logical_name}SecurityGroup",
                GroupName=Sub(
                    f"${{StackName}}-{cluster.logical_name}",
                    StackName=define_stack_name(),
                ),
                GroupDescription=Sub(
                    f"${{StackName}} {cluster.name}", StackName=define_stack_name()
                ),
                VpcId=Ref(VPC_ID),
            ),
        )
        cluster_sg_id = GetAtt(
            cluster.security_group, MSK_CLUSTER_SG_PARAM.return_value
        )
        cluster_props = import_record_properties(
            cluster.properties,
            AwsMskCluster,
        )
        broker_info = set_else_none("BrokerNodeGroupInfo", cluster_props, {})
        security_groups = set_else_none("SecurityGroups", broker_info, [cluster_sg_id])
        if security_groups and cluster_sg_id not in security_groups:
            security_groups.append(
                GetAtt(cluster.security_group, MSK_CLUSTER_SG_PARAM.return_value)
            )
        if broker_info:
            setattr(broker_info, "SecurityGroups", security_groups)
            client_subnets = (
                getattr(broker_info, "ClientSubnets")
                if hasattr(broker_info, "ClientSubnets")
                else []
            )
            if not client_subnets:
                setattr(broker_info, "ClientSubnets", Ref(cluster.subnets_override))
        else:
            cluster_props.update(
                {
                    "BrokerNodeGroupInfo": {"SecurityGroups": security_groups},
                    "ClientSubnets": Ref(cluster.subnets_override),
                }
            )
        cluster.cfn_resource = AwsMskCluster(cluster.logical_name, **cluster_props)
        version_param = cluster_template.add_parameter(
            validate_cluster_version(cluster, cluster.cfn_resource.KafkaVersion)
        )
        setattr(cluster.cfn_resource, "KafkaVersion", Ref(version_param))
        cluster_stack = MskClusterStack(
            cluster.logical_name,
            cluster=cluster,
            stack_template=cluster_template,
            top_stack=msk_top_stack,
            stack_parameters={
                ROOT_STACK_NAME.title: Ref(ROOT_STACK_NAME),
                VPC_ID.title: Ref(VPC_ID),
                STORAGE_SUBNETS.title: Ref(STORAGE_SUBNETS),
                PUBLIC_SUBNETS.title: Ref(PUBLIC_SUBNETS),
                APP_SUBNETS.title: Ref(APP_SUBNETS),
                version_param.title: version_param.Default,
            },
        )
        set_instance_type(cluster, cluster_stack)

        handle_msk_auth_settings(cluster)
        add_resource(cluster_template, cluster.cfn_resource)
        add_storage_scaling(cluster, cluster_template)
        add_resource(msk_top_stack.stack_template, cluster_stack)
        cluster.init_outputs()
        cluster.generate_outputs()
        add_outputs(cluster_template, cluster.outputs)

        new_outputs = []
        for output_name in cluster.stack.stack_template.outputs:
            new_outputs.append(
                Output(
                    output_name,
                    Value=GetAtt(cluster_stack.title, f"Outputs.{output_name}"),
                )
            )
        add_outputs(msk_top_stack.stack_template, new_outputs)



#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>

from os import path
from pathlib import Path

from ecs_composex.mods_manager import XResourceModule

from .msk_cluster import MskCluster
from .msk_cluster_stack import XStack

COMPOSE_X_MODULES: dict = {
    "x-msk_cluster": {
        "Module": XResourceModule(
            "x-msk_cluster",
            XStack,
            Path(path.abspath(path.dirname(__file__))),
            MskCluster,
        ),
    }
}



#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>

"""
Manages the IAM Permissions & SCRAM configuration for authentication to the Kafka cluster resources
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ecs_composex.common.settings import ComposeXSettings
    from ecs_composex.ecs.ecs_family import ComposeFamily
    from ecs_composex_msk_cluster.msk_cluster import MskCluster

from compose_x_common.compose_x_common import set_else_none
from ecs_composex.common.logging import LOG
from ecs_composex.common.troposphere_tools import add_resource
from troposphere import Ref, Select, Split, Sub
from troposphere.iam import PolicyType

from .msk_cluster_params import MSK_CLUSTER_ARN

MSK_TOPIC_PERMS_MAPPING: dict = {
    "Admin": [
        "kafka-cluster:*Topic*",
        "kafka-cluster:ReadData",
        "kafka-cluster:WriteData",
    ],
    "Producer": ["kafka-cluster:DescribeTopic", "kafka-cluster:WriteData"],
    "Consumer": ["kafka-cluster:DescribeTopic", "kafka-cluster:ReadData"],
    "ProducerConsumer": [
        "kafka-cluster:DescribeTopic",
        "kafka-cluster:WriteData",
        "kafka-cluster:ReadData",
    ],
}
MSK_GROUP_PERMS_MAPPING: dict = {
    "Admin": [
        "kafka-cluster:DescribeGroup",
        "kafka-cluster:AlterGroup",
        "kafka-cluster:DeleteGroup",
    ],
    "Consumer": [
        "kafka-cluster:DescribeGroup",
        "kafka-cluster:AlterGroup",
    ],
}
MSK_TRANSACTIONAL_IDS_PERMS_MAPPING: dict = {
    "Producer": [
        "kafka-cluster:AlterTransactionalId",
        "kafka-cluster:DescribeTransactionalId",
    ],
}


def handle_topic_iam_policy(
    resource: MskCluster,
    topics_def: dict,
    statement: list,
    cluster_arn_prefix: Select,
) -> None:
    """
    Creates the IAM policies for MSK IAM connectivity to Kafka Topics
    """
    for access_type, topics_regex in topics_def.items():
        if access_type not in MSK_TOPIC_PERMS_MAPPING:
            LOG.warning(
                f"{resource.module.res_key}.{resource.name}"
                " Topic permissions {access_type} is invalid."
                f" Must be one of {list(MSK_TOPIC_PERMS_MAPPING.keys())}"
            )
            continue
        access_type_statement: dict = {
            "Sid": f"Topic{access_type}",
            "Effect": "Allow",
            "Action": MSK_TOPIC_PERMS_MAPPING[access_type],
            "Resource": [
                Sub(
                    f"${{CLUSTER_ARN_PREFIX}}:topic/${{CLUSTER_ID}}/{regex}",
                    CLUSTER_ARN_PREFIX=cluster_arn_prefix,
                    CLUSTER_ID=resource.cluster_uuid,
                )
                for regex in topics_regex
            ],
        }
        statement.append(access_type_statement)


def handle_group_iam_policy(
    resource: MskCluster,
    groups_def: dict,
    statement: list,
    cluster_arn_prefix: Select,
) -> None:
    """
    Creates the IAM policies for MSK IAM connectivity to Kafka Topics
    """
    for access_type, topics_regex in groups_def.items():
        if access_type not in MSK_GROUP_PERMS_MAPPING:
            LOG.warning(
                f"{resource.module.res_key}.{resource.name}"
                " Topic permissions {access_type} is invalid."
                f" Must be one of {list(MSK_GROUP_PERMS_MAPPING.keys())}"
            )
            continue

        access_type_statement: dict = {
            "Sid": f"Group{access_type}",
            "Effect": "Allow",
            "Action": MSK_GROUP_PERMS_MAPPING[access_type],
            "Resource": [
                Sub(
                    f"${{CLUSTER_ARN_PREFIX}}:group/${{CLUSTER_ID}}/{regex}",
                    CLUSTER_ARN_PREFIX=cluster_arn_prefix,
                    CLUSTER_ID=resource.cluster_uuid,
                )
                for regex in topics_regex
            ],
        }
        statement.append(access_type_statement)


def handle_transactional_iam_policy(
    resource: MskCluster,
    transactional_def: dict,
    statement: list,
    cluster_arn_prefix: Select,
) -> None:
    """
    Creates the IAM policies for MSK IAM connectivity to Kafka Topics
    """
    for access_type, topics_regex in transactional_def.items():
        if access_type not in MSK_TRANSACTIONAL_IDS_PERMS_MAPPING:
            LOG.warning(
                f"{resource.module.res_key}.{resource.name}"
                " Transactional ID permissions {access_type} is invalid."
                f" Must be one of {list(MSK_TRANSACTIONAL_IDS_PERMS_MAPPING.keys())}"
            )
            continue
        access_type_statement: dict = {
            "Sid": f"Transactions{access_type}",
            "Effect": "Allow",
            "Action": MSK_TRANSACTIONAL_IDS_PERMS_MAPPING[access_type],
            "Resource": [
                Sub(
                    f"${{CLUSTER_ARN_PREFIX}}:transactional-id/${{CLUSTER_ID}}/{regex}",
                    CLUSTER_ARN_PREFIX=cluster_arn_prefix,
                    CLUSTER_ID=resource.cluster_uuid,
                )
                for regex in topics_regex
            ],
        }
        statement.append(access_type_statement)


def handle_cluster_permissions(
    resource: MskCluster,
    settings: ComposeXSettings,
    cluster_def: dict,
    statement: list,
    cluster_arn_prefix,
) -> None:
    """
    Handle Cluster permissions
    """
    idempotent_write = set_else_none("IdempotentWrite", cluster_def, False)
    if idempotent_write:
        statement.append(
            {
                "Effect": "Allow",
                "Action": ["kafka-cluster:WriteDataIdempotently"],
                "Sid": "ClusterWriteIdempotently",
                "Resource": Ref(resource.cluster_arn_parameter),
            }
        )


def handle_iam_kafka_resources_access(
    resource: MskCluster,
    settings: ComposeXSettings,
    family: ComposeFamily,
    family_target_def: dict,
) -> None:
    """
    Creates IAM Policies for access to MSK resources [topic, group, cluster, transactional-id]
    """
    if resource.cfn_resource:
        cluster_arn_id = resource.add_attribute_to_another_stack(
            family.stack, resource.cluster_arn_parameter, settings
        )
        cluster_arn_prefix = Select(
            0, Split(":cluster/", Ref(cluster_arn_id["ImportParameter"]))
        )
    else:
        cluster_arn_id = resource.attributes_outputs[MSK_CLUSTER_ARN]["ImportValue"]
        cluster_arn_prefix = Select(0, Split(":cluster/", cluster_arn_id))

    statement: list = [
        {
            "Sid": "ConnectToCluster",
            "Effect": "Allow",
            "Resource": [
                (
                    Ref(cluster_arn_id["ImportParameter"])
                    if resource.cfn_resource
                    else cluster_arn_id
                )
            ],
            "Action": ["kafka-cluster:Connect"],
        }
    ]
    handle_cluster_permissions(
        resource,
        settings,
        set_else_none("cluster", family_target_def, {}),
        statement,
        cluster_arn_prefix,
    )
    handle_topic_iam_policy(
        resource,
        set_else_none("topic", family_target_def, {}),
        statement,
        cluster_arn_prefix,
    )
    handle_group_iam_policy(
        resource,
        set_else_none("group", family_target_def, {}),
        statement,
        cluster_arn_prefix,
    )
    handle_transactional_iam_policy(
        resource,
        set_else_none("transactional-id", family_target_def, {}),
        statement,
        cluster_arn_prefix,
    )

    iam_policy = PolicyType(
        f"KafkaAccess{family.logical_name}{resource.module.mapping_key}{resource.logical_name}",
        PolicyName=f"KafkaAccess{family.logical_name}{resource.module.mapping_key}{resource.logical_name}",
        PolicyDocument={"Version": "2012-10-17", "Statement": statement},
        Roles=[family.iam_manager.task_role.name],
    )
    add_resource(family.template, iam_policy)



# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2022 John Mille <john@compose-x.io>

"""Main module."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ecs_composex.mods_manager import XResourceModule
    from ecs_composex.common.settings import ComposeXSettings

from compose_x_common.compose_x_common import keyisset
from ecs_composex.common.stacks import ComposeXStack
from ecs_composex.common.troposphere_tools import build_template
from ecs_composex.vpc.vpc_params import (
    APP_SUBNETS,
    PUBLIC_SUBNETS,
    STORAGE_SUBNETS,
    VPC_ID,
)

from .msk_cluster import define_msk_clusters_mappings
from .msk_cluster_template import build_msk_clusters


class XStack(ComposeXStack):
    """
    Class to handle MSK resources
    """

    def __init__(
        self, title: str, settings: ComposeXSettings, module: XResourceModule, **kwargs
    ):
        if module.lookup_resources:
            if not keyisset(module.mapping_key, settings.mappings):
                settings.mappings[module.mapping_key] = {}
            define_msk_clusters_mappings(module, settings)
        if module.new_resources:
            stack_template = build_template(
                f"{module.res_key} - Root Stack",
                [
                    VPC_ID,
                    STORAGE_SUBNETS,
                    APP_SUBNETS,
                    PUBLIC_SUBNETS,
                ],
            )
            super().__init__(title, stack_template, **kwargs)
            build_msk_clusters(module, self)
            self.parent_stack = settings.root_stack
        else:
            self.is_void = True
        for resource in module.lookup_resources:
            resource.stack = self



#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>

from ecs_composex.common.cfn_params import Parameter
from ecs_composex.vpc.vpc_params import SG_ID_TYPE

GROUP_LABELS: str = "MSK Settings"

MSK_CLUSTER_INSTANCE_TYPES_T = "MskInstanceType"
MSK_CLUSTER_INSTANCE_TYPES = Parameter(
    MSK_CLUSTER_INSTANCE_TYPES_T,
    Type="String",
    AllowedValues=[
        "kafka.t3.small",
        "kafka.m5.large",
        "kafka.m5.xlarge",
        "kafka.m5.2xlarge",
        "kafka.m5.4xlarge",
        "kafka.m5.8xlarge",
        "kafka.m5.12xlarge",
        "kafka.m5.16xlarge",
        "kafka.m5.24xlarge",
    ],
    Default="kafka.m5.large",
)

MSK_CLUSTER_SG_PARAM_T = "MSKClusterSecurityGroup"
MSK_CLUSTER_SG_PARAM = Parameter(
    MSK_CLUSTER_SG_PARAM_T,
    group_label=GROUP_LABELS,
    return_value="GroupId",
    Type=SG_ID_TYPE,
)

MSK_CLUSTER_ARN_T = "ClusterArn"
MSK_CLUSTER_ARN = Parameter(MSK_CLUSTER_ARN_T, group_label=GROUP_LABELS, Type="String")
MSK_CLUSTER_CLIENTS_SHARED_SG_T = "MSKClientsSecurityGroup"
MSK_CLUSTER_CLIENTS_SHARED_SG = Parameter(
    MSK_CLUSTER_CLIENTS_SHARED_SG_T,
    return_value="GroupId",
    group_label=GROUP_LABELS,
    Type=SG_ID_TYPE,
    Description="MSK Clients Security Group",
)

MSK_CLUSTER_ADDRESSING_TYPE_T = "MskClusterNetworkAccessType"
MSK_CLUSTER_ADDRESSING_TYPE = Parameter(
    MSK_CLUSTER_ADDRESSING_TYPE_T,
    group_label=GROUP_LABELS,
    Type="String",
    AllowedValues=["PUBLIC", "PRIVATE"],
    Default="PRIVATE",
    Description="MSK Cluster public access is enabled",
)

MSK_CLUSTER_USE_SASL_IAM_T = "UseSaslIam"
MSK_CLUSTER_USE_SASL_IAM = Parameter(
    MSK_CLUSTER_USE_SASL_IAM_T,
    group_label=GROUP_LABELS,
    Type="String",
    AllowedValues=["True", "False"],
    Default="True",
    Description="Use IAM for SASL Authentication",
)
MSK_CLUSTER_USE_SASL_SCRAM_T = "UseSaslScram"
MSK_CLUSTER_USE_SASL_SCRAM = Parameter(
    MSK_CLUSTER_USE_SASL_SCRAM_T,
    group_label=GROUP_LABELS,
    Type="String",
    AllowedValues=["True", "False"],
    Default="True",
    Description="Use SCRAM for SASL Authentication",
)

MSK_CLUSTER_USE_TLS_AUTH_T = "UseSaslScram"
MSK_CLUSTER_USE_TLS_AUTH = Parameter(
    MSK_CLUSTER_USE_TLS_AUTH_T,
    group_label=GROUP_LABELS,
    Type="String",
    AllowedValues=["True", "False"],
    Default="False",
    Description="Use TLS for Authentication",
)

MSK_PORTS_MAPPING: dict = {
    "Private": {
        "Plaintext": 9092,
        "Tls": 9094,
        "Scram": 9096,
        "Iam": 9098,
    },
    "Public": {
        "Plaintext": 9192,
        "Tls": 9194,
        "Scram": 9096,
        "Iam": 9198,
    },
    "Zookeeper": {"Plaintext": 2181, "Tls": 2182},
}

CONTROL_CLOUD_ATTR_MAPPING = {
    MSK_CLUSTER_ARN: "Arn",
    MSK_CLUSTER_SG_PARAM: "BrokerNodeGroupInfo::SecurityGroups::0",
}
